__description__ = "a module that houses methods and classes used to produce html reports"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import os
import json
from collections import defaultdict
import numpy as np

import time
import getpass ### used to figure out who ran this

try:
    from html import HTML
except ImportError:
    HTML = None

### dependencies not in standard Python libraries
from idq import configparser
from idq import plots
from idq import names
from idq import features
from idq import factories
from idq import logs
from idq import utils
from idq import io

from ligo import segments as ligo_segments

from idq import __version__ ### import this for the "entire library"
__process_name__ = "iDQ"

#-------------------------------------------------

DEFAULT_PGLITCH_THRESHOLD = 0.9
DEFAULT_FIGTYPE = 'png'
DEFAULT_LINKS = [
    ('https://git.ligo.org/reed.essick/iDQ', 'iDQ source code'),
    ('https://git.ligo.org/reed.essick/iDQ/wikis', 'iDQ wiki'),
    ('https://docs.ligo.org/reed.essick/iDQ/', 'iDQ docs'),
]

#-------------------------------------------------

class Report(object):
    """generate everything needed for a DQR report
    """
    _title_fontsize = 14

    def __init__(self, config_path, start, end, t0=None, zoom_start=None, zoom_end=None):
        self._start = start
        self._end = end
        self._t0 = t0
        self._zoom_start = zoom_start
        self._zoom_end = zoom_end
        self._path2config(config_path)

    @property
    def config(self):
        return self._config

    @property
    def config_path(self):
        return self._config_path

    @property
    def instrument(self):
        return self._instrument

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def zoom_start(self):
        return self._zoom_start if (self._zoom_start is not None) else self._start

    @property
    def zoom_end(self):
        return self._zoom_end if (self._zoom_end is not None) else self._end

    @property
    def t0(self):
        return self._t0 if (self._t0 is not None) else 0.5*(self.zoom_start+self.zoom_end)

    @property
    def tag(self):
        return self._tag

    @property
    def classifier_data(self):
        return self._classifier_data

    @property
    def quiver_factory(self):
        return factories.QuiverFactory(self.classifier_data)

    def _path2config(self, config_path):
        self._config_path = os.path.abspath(os.path.realpath(config_path))
        self._config = configparser.path2config(config_path)
        self._instrument = self.config.get('general', 'instrument', '')
        self._tag = self.config.get('general', 'tag')

        target_bounds = configparser.config2bounds(self.config.get('samples', 'target_bounds'))
        frequency = self.config.get('timeseries data discovery', 'frequency')
        assert target_bounds.has_key(frequency), 'must specify a frequency range (called %s) within target_bounds'%frequency
        freq_min, freq_max = target_bounds[frequency]
        assert isinstance(freq_min, int) and isinstance(freq_max, int), 'frequency bounds must be integers!'
        self._freq_bounds = freq_min, freq_max 

        rootdir = os.path.abspath(os.path.realpath(self.config.get('general','rootdir')))
        timeseriesdir = names.tag2timeseriesdir(self.tag, rootdir=rootdir)
        traindir = names.tag2traindir(self.tag, rootdir=rootdir)
        evaluatedir = names.tag2evaluatedir(self.tag, rootdir=rootdir)
        calibratedir = names.tag2calibratedir(self.tag, rootdir=rootdir)

        ### reporters
        reporter_factory = factories.ReporterFactory()

        # timeseries
        ritems = configparser.evalkwargs(**dict(self.config.items('timeseries reporting')))
        self.timeseriesreporter = reporter_factory(timeseriesdir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        # train
        ritems = configparser.evalkwargs(**dict(self.config.items('train reporting')))
        self.trainreporter = reporter_factory(traindir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        # evaluate
        ritems = configparser.evalkwargs(**dict(self.config.items('evaluate reporting')))
        self.evaluatereporter = reporter_factory(evaluatedir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        # calibrate
        ritems = configparser.evalkwargs(**dict(self.config.items('calibrate reporting')))
        self.calibratereporter = reporter_factory(calibratedir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        ### classifier data
        classifier_data_factory = factories.ClassifierDataFactory()
        citems = configparser.evalkwargs(**dict(self.config.items('evaluate data discovery')))
        self._classifier_data = classifier_data_factory(self.zoom_start, self.zoom_end, **configparser.add_missing_kwargs(citems, group=names.tag2group(self.tag, 'timeseries')))

    def _filter_series(self, series):
        start = series['t0']
        dt = series['deltaT']
        end = start + dt*len(series['data'].values()[0])
        times = np.arange(start, end, dt)
        series['data']['times'] = times

        if start < self.start: ### truncate the start
            truth = series['data']['times']>=self.start
            for key, val in series['data'].items():
                series['data'][key] = val[truth]
            series['t0'] = self.start

        if end > self.end: ### turncate the end
            truth = series['data']['times']<self.end
            for key, val in series['data'].items():
                series['data'][key] = val[truth]

        return series

    def _find_model_segs(self, series_dict, logger=None):
        if logger is not None:
            logger.info('finding spans used for models within [%.3f, %.3f)'%(self.start, self.end))
        segdict = {nickname: defaultdict(ligo_segments.segmentlist) for nickname in series_dict.keys()}
        for nickname, series in series_dict.items():
            for s in series:
                seg = ligo_segments.segment(s['t0'], s['t0']+ s['data']['times'].size * s['deltaT'])
                segdict[nickname][s['model']].append(seg)
        for nickname in segdict.keys():
            for model in segdict[nickname].keys():
                segdict[nickname][model].coalesce()
        return segdict

    def _glob_timeseries(self, nicknames, logger=None):
        if logger is not None:
            logger.info('globbing timeseries within [%.3f, %.3f)'%(self.start, self.end))
        series = defaultdict(list)
        for nickname in nicknames:
            for _, _, S in self.timeseriesreporter.glob(names.nickname2timeseries_topic(self.instrument, nickname), self.start, self.end):
                for s in S:
                    s = self._filter_series(s)
                    if len(s['data']['times']): ### some remaining time
                        series[nickname].append(s)
        if logger is not None:
            for nickname in nicknames:
                logger.info('found %d series for %s'%(len(series[nickname]), nickname))
        return series

    def _retrieve_timeseries(self, hashes_dict, logger=None):
        if logger is not None:
            logger.info('retrieving timeseries within [%.3f, %.3f) based on specified hashes'%(self.start, self.end))
        series = defaultdict(list)
        for nickname, hashes in hashes_dict.items():
            name = names.nickname2timeseries_topic(self.instrument, nickname)
            for h in hashes:
                self.timeseriesreporter.start, self.timeseriesreporter.end = names.hash2start_end(h)
                for s in self.timeseriesreporter.retrieve(name):
                    s = self._filter_series(s)
                    if len(s['data']['times']): ### some remaining time
                        series[nickname].append(s)
        if logger is not None:
            for nickname, hashes in hashes_dict.items():
                logger.info('found %d series out of %d requested for %s'%(len(series[nickname]), len(hashes), nickname))
        return series

    def _glob_models(self, nicknames, logger=None):
        if logger is not None:
            logger.info('globbing models within [%.3f, %.3f)'%(self.start, self.end))
        models = defaultdict(dict)
        for nickname in nicknames:
            for s, e, m in self.trainreporter.glob(names.nickname2train_topic(nickname), self.start, self.end):
                models[nickname].update({names.start_end2hash(s, e):m})
        if logger is not None:
            for nickname in nicknames:
                logger.info('found %d models for %s'%(len(models[nickname]), nickname))
        return models

    def  _retrieve_models(self, hashes_dict, logger=None):
        if logger is not None:
            logger.info('retrieving models within [%.3f, %.3f) base on specified hashes'%(self.start, self.end))
        models = defaultdict(dict)
        for nickname, hashes in hashes_dict.items():
            name = names.nickname2train_topic(nickname)
            for h in hashes:
                self.trainreporter.start, self.trainreporter.end = names.hash2start_end(h)
                models[nickname].update({h:self.trainreporter.retrieve(name)})
        if logger is not None:
            for nickname, hashes in hashes_dict.items():
                logger.info('found %d models out of %d requested for %s'%(len(models[nickname]), len(hashes), nickname))
        return models

    def _glob_calibration_maps(self, nicknames, logger=None):
        if logger is not None:
            logger.info('globbing calibration maps within [%.3f, %.3f)'%(self.start, self.end))
        calibs = defaultdict(dict)
        for nickname in nicknames:
            for _, _, c in self.calibratereporter.glob(names.nickname2calibrate_topic(nickname), self.start, self.end):
                calibs[nickname].update({names.start_end2hash(s,e):c})
        if logger is not None:
            for nickname in nicknames:
                logger.info('found %d calibration maps for %s'%(len(calibs[nickname]), nickname))
        return calibs

    def _retrieve_calibration_maps(self, hashes_dict, logger=None):
        if logger is not None:
            logger.info('retrieving calibration maps within [%.3f, %.3f) based on specified hashes'%(self.start, self.end))
        calibs = defaultdict(dict)
        for nickname, hashes in hashes_dict.items():
            name = names.nickname2calibrate_topic(nickname)
            for h in hashes:
                self.calibratereporter.start, self.calibratereporter.end = names.hash2start_end(h)
                calibs[nickname].update({h:self.calibratereporter.retrieve(name)})
        if logger is not None:
            for nickname, hashes in hashes_dict.items():
                logger.info('found %d calibration maps out of %d requested for %s'%(len(calibs[nickname]), len(hashes), nickname))
        return calibs

    def _glob_quivers(self, nicknames, logger=None):
        if logger is not None:
            logger.info('globbing evaluated quivers within [%.3f, %.3f)'%(self.start, self.end))
        quivers = dict((nickname, features.Quiver()) for nickname in nicknames)
        segs = ligo_segments.segmentlist([ligo_segments.segment(self.start, self.end)])
        for nickname in nicknames:
            for _, _, q in self.evaluatereporter.glob(names.nickname2evaluate_topic(nickname), self.start, self.end):
                q.filter(segs=utils.segments_intersection(q.segs, segs)) ### do this so we don't mess up quiver's internals
                quivers[nickname].append_quiver(q)
        if logger is not None:
            for nickname in nicknames:
                gch, cln = quivers[nickname].vectors2classes()
                logger.info('found %d gch samples and %d cln samples for %s'%(len(gch), len(cln), nickname))
        return quivers

    def path(self, name, output_dir='.', suffix='png'):
        return os.path.join(output_dir, '%s-%s.%s'%(name, self.instrument+__process_name__+self.tag, suffix))

    def _stats(self,
            nicknames,
            quivers,
            models,
            calibmaps,
            series=None,
            logger=None,
            pglitch_thr=DEFAULT_PGLITCH_THRESHOLD,
        ):
        if logger is not None:
            logger.info('computing basic statistics')
        cols = [
            'nickname',
            'number of glitch samples',
            'number of clean samples',
            'number of models',
            'number of calibration maps',
        ]
        if series is not None:
            cols.append('% time with p(glitch|aux)>='+'%.3f'%pglitch_thr)

        data = []
        for nickname in nicknames:
            # do some quick data manipulations
            gch, cln = quivers[nickname].vectors2classes()
            datum = [
                nickname,
                len(gch),
                len(cln),
                len(models[nickname]),
                len(calibmaps[nickname]),
            ]

            if series is not None:
                key = names.channel_name_template(self._instrument, nickname, *self._freq_bounds)%names.PGLITCH
                times = np.concatenate(tuple(s['data']['times'] for s in series[nickname]))
                pglitch = np.concatenate(tuple(s['data'][key] for s in series[nickname]))
                pglitch = pglitch[(self.zoom_start<=times)*(times<self.zoom_end)]
                datum.append(100.*np.sum(pglitch>=pglitch_thr)/len(pglitch))

            data.append(datum)

        return cols, data

    def _timeseries_figure(self, nicknames, series, gch_gps=[], legend=True, segs=None, output_dir='.', figtype=DEFAULT_FIGTYPE, logger=None):
        if logger is not None:
            logger.info('generating timeseries plots')
        alttext = 'timeseries'

        fig = plots.timeseries(nicknames, series, self.zoom_start, self.zoom_end, self.t0, gch_gps=gch_gps, legend=legend, segs=segs)
        fig.suptitle('%s Timeseries\nwithin [%.3f, %.3f)'%(self.instrument, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(alttext, output_dir=output_dir, suffix=figtype)

        if logger is not None:
            logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)
        return alttext, path

    def _calibration_coverage(self, nicknames, series, quivers, legend=True, output_dir='.', figtype=DEFAULT_FIGTYPE, logger=None):
        if logger is not None:
            logger.info('generating calibration coverage plots')
        alttext = 'calibration_coverage'

        fig = plots.calibration_coverage(nicknames, series, quivers, legend=legend)
        fig.suptitle('%s Calibration Coverage\nwithin [%.3f, %.3f)'%(self.instrument, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(alttext, output_dir=output_dir, suffix=figtype)

        if logger is not None:
            logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)
        return alttext, path

    def _calibration_distribs(self, nicknames, quivers, calibmaps, legend=True, output_dir='.', figtype=DEFAULT_FIGTYPE, logger=None):
        if logger is not None:
            logger.info('generating calibration distribution plots')
        alttext = 'calibration_distribs'

        fig = plots.calibration_distribs(nicknames, quivers, calibmaps, legend=legend)
        fig.suptitle('%s Calibration Distributions\nwithin [%.3f, %.3f)'%(self.instrument, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(alttext, output_dir=output_dir, suffix=figtype)

        if logger is not None:
            logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)
        return alttext, path

    def _roc(self, nicknames, quivers, calibmaps, series=None, legend=True, output_dir='.', figtype=DEFAULT_FIGTYPE, logger=None):
        if logger is not None:
            logger.info('generating ROC plots')
        alttext = 'roc'

        fig = plots.roc(nicknames, series, quivers, calibmaps, legend=legend)
        fig.suptitle('%s Receiver Operating Characteristics\nwithin [%.3f, %.3f)'%(self.instrument, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(alttext, output_dir=output_dir, suffix=figtype)

        if logger is not None:
            logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)
        return alttext, path

    def _feature_importance(self, nicknames, models, segdict, legend=True, output_dir='.', figtype=DEFAULT_FIGTYPE, logger=None):
        if logger is not None:
            logger.info('generating feature importance plots and tables')

        # instantiate classifiers to pick up some standard stuff...
        classifier_factory = factories.ClassifierFactory()
        klassifiers = dict((nickname, classifier_factory(nickname, rootdir=output_dir, **configparser.evalkwargs(**dict(self.config.items(nickname))))) for nickname in nicknames)

        # actually compute stuff
        figs, tabs = plots.featureimportance(nicknames, models, klassifiers, segdict, self.quiver_factory, self.zoom_start, self.zoom_end, self.t0, legend=legend)
        paths = []
        cols_datas = []
        for nickname in nicknames:
            for key, model in models[nickname].items():
                fig = figs[nickname][key]
                tab = tabs[nickname][key]
                if (fig is None) and (tab is None) and (logger is not None):
                    logger.warn('feature importance not available for nickname='+nickname)

                if fig is not None:
                    alttext = nickname+'-'+self.instrument+'_featureimportance-'+model.hash
                    path = self.path(alttext, output_dir=output_dir, suffix=figtype)
                    fig.suptitle('%s %s Feature Importance\nwithin [%.3f, %.3f)'%(self.instrument, nickname, self.start, self.end), fontsize=self._title_fontsize)
                    if logger is not None:
                        logger.info('saving: '+path)
                    fig.savefig(path)
                    plots.close(fig)
                    paths.append((alttext, path))

                if tab is not None:
                    cols, data = tab
                    model_name = " ".join([nickname, model.hash])
                    cols_datas.append((model_name, cols, data))

        return paths, cols_datas

    def report(self,
            nicknames,
            reportdir='.',
            logdir='.',
            verbose=False,
            figtype=DEFAULT_FIGTYPE,
            links=DEFAULT_LINKS,
            ignore_segdb=False,
            skip_timeseries=False,
            annotate_gch_gps=False,
            pglitch_thr=DEFAULT_PGLITCH_THRESHOLD,
            legend=True,
            log_level=logs.DEFAULT_LOG_LEVEL,
        ):
        ### act like a DiskReporter...
        reportdir = names.start_end2fragmented_dir(self.start, self.end, rootdir=reportdir)
        if not os.path.exists(reportdir):
            os.makedirs(reportdir)

        logger = logs.get_logger(
            names.tag2report_logname(self.tag),
            log_level=log_level,
            rootdir=logdir,
            verbose=verbose
        )
        logger.info('generating report for %s within [%.3f, %.3f) with %s'%(', '.join(nicknames), self.start, self.end, self.config_path))

        ### go get all the data you'll need
        # evaluated quivers
        quivers = self._glob_quivers(nicknames, logger=logger)

        # models and calibmaps
        if skip_timeseries: ### just grab everything available
            series = None
            models = self._glob_models(nicknames, logger=logger)
            calibmaps = self._glob_calibration_maps(nicknames, logger=logger)
            segdict = None
        else: ### grab series and only load the ones used to generate this data
            series = self._glob_timeseries(nicknames, logger=logger)
            models = self._retrieve_models(dict((nickname, set(s['model'] for s in series[nickname])) for nickname in nicknames), logger=logger)
            calibmaps = self._retrieve_calibration_maps(dict((nickname, set(s['calib'] for s in series[nickname])) for nickname in nicknames), logger=logger)
            segdict = self._find_model_segs(series, logger=logger)

        if ignore_segdb:
            logger.info( "ignoring segdb" )
            segs = ligo_segments.segmentlist([ligo_segments.segment(self.start, self.end)])

        else:
            segdb_intersect = self.config.get('segments', 'intersect').split() if self.config.has_option('segments', 'intersect') else []
            segdb_exclude = self.config.get('segments', 'exclude').split() if self.config.has_option('segments', 'exclude') else []
            segdb_url = self.config.get('segments', 'segdb_url')

            logger.info( "querying %s within [%.3f, %.3f) for: intersect=%s ; exclude=%s"%(segdb_url, self.start, self.end, ','.join(segdb_intersect), ','.join(segdb_exclude)) )

            segs = utils.segdb2segs(
                self.start,
                self.end,
                intersect=segdb_intersect,
                exclude=segdb_exclude,
                segdb_url=segdb_url,
                logger=logger,
            )
        lvtm = utils.livetime(segs)
        logger.info('retained %.3f sec of livetime'%lvtm)

        ### record segments
        ### do this backflip so segment reporter doesn't create subdirectories...
        segmentreporter = io.SegmentReporter(reportdir, self.start, self.end)
        path = os.path.join(reportdir, names.start_end2path('segments', self.start, self.end, suffix=segmentreporter.suffix))
        segmentreporter._write(path, segs)
        logger.info( 'segments written to '+path )

        ### generate report as an html document
        generate_html = HTML is not None ### build a doc if we have the library available

        # add banner header
        if generate_html:
            doc, _, body = self.html_head_body()
            body.br()
            div = body.div(klass='jumbotron', id='Banner')
            div.h1('%s Report for %s (%s)'%(\
                __process_name__,
                self.instrument,
                self.tag,
            ))
            div.br()
            table = div.table(klass="table table-condensed table-hover")
            thead = table.thead(style='display:none;')
            thead.th('type')
            thead.th('info')
            tbody = table.tbody()
            tr = tbody.tr()
            tr.th('classifiers', scope='row')
            tr.td(', '.join(nicknames))
            tr = tbody.tr()
            tr.th('span', scope='row')
            tr.td('[%.3f, %.3f)'%(self.start, self.end))
            tr = tbody.tr()
            tr.th('livetime', scope='row')
            tr.td('%.3f seconds'%lvtm)
            tr = tbody.tr()
            tr.th('duty cycle', scope='row')
            tr.td('%.3f'%(lvtm/(self.end-self.start)))

        ### generate a table summarizing the basic numbers for each nickname
        args = self._stats(
            nicknames,
            quivers,
            models,
            calibmaps,
            series=series,
            logger=logger,
            pglitch_thr=pglitch_thr,
        )
        if generate_html:
            body.hr()
            div = body.div(klass='container', id='BasicStatistics')
            div.h2('Basic Statistics')
            div.br()
            self._add_table(div, *args)
            div.br()

        ### make plots
        # ROC curve
        args = self._roc(
            nicknames,
            quivers,
            calibmaps,
            series=series,
            legend=legend,
            output_dir=reportdir,
            figtype=figtype,
            logger=logger,
        )
        if generate_html:
            body.hr()
            div = body.div(klass='container', id='ROC')
            div.h2('Receiver Operating Characteristic Curves')
            div.br()
            self._add_figure(div, *args)
            div.br()

        # series and calib stuff that depends on series
        if series is not None:
            gch_gps = []
            if annotate_gch_gps:
                for q in quivers.values():
                    gch_gps += list(q.vectors2classes()[0].gpss)
                gch_gps = sorted(set(gch_gps))

            # timeseries
            args = self._timeseries_figure(
                nicknames,
                series,
                gch_gps=gch_gps,
                legend=legend,
                segs=segs,
                output_dir=reportdir,
                figtype=figtype,
                logger=logger,
            )
            if generate_html:
                body.hr()
                div = body.div(klass='container', id='Timeseries')
                div.h2('Timeseries')
                div.br()
                self._add_figure(div, *args)
                div.br()

            # calibration coverage
            args = self._calibration_coverage(
                nicknames,
                series,
                quivers,
                legend=legend,
                output_dir=reportdir,
                figtype=figtype,
                logger=logger,
            )
            if generate_html:
                body.hr()
                div = body.div(klass='container', id='Calibration') ### will be used below as well
                div.h2('Calibration')
                div.br()
                self._add_figure(div, *args)
                div.br()

        elif generate_html:
            div = body.div(klass='container', id='Calibration') ### will be used below

        # calibration distributions
        args = self._calibration_distribs(
            nicknames,
            quivers,
            calibmaps,
            legend=legend,
            output_dir=reportdir,
            figtype=figtype,
            logger=logger,
        )
        if generate_html:
            self._add_figure(div, *args)
            div.br()

        # feature importance
        importance_paths, importance_cols_datas = self._feature_importance(
            nicknames,
            models,
            segdict,
            legend=legend,
            output_dir=reportdir,
            figtype=figtype,
            logger=logger,
        )

        if generate_html:
            body.hr()
            div = body.div(klass='container', id='FeatureImportance')
            div.h2('Feature Importance')
            div.br()
            for alttext, path in importance_paths:
                self._add_figure(div, alttext, path)
            div.br()

            div = body.div(klass='container', id='FeatureImportanceTables')
            for model_name, cols, data in importance_cols_datas:
                div.h4('model: %s'%model_name )
                self._add_table(div, cols, data)
            div.br()

            # write date into a footer
            div = body.div(klass='container', id='footer')
            div.p('generated at %s'%time.strftime("%H:%M:%S %Z %a %d %b %Y", time.localtime()))
 
            ### write html page
            path = self.write_html(doc, reportdir=reportdir)
            logger.info('html document written to: '+path)

    def write_html(self, doc, reportdir):
        path = self.path('report', output_dir=reportdir, suffix='html')
        with open(path, 'w') as obj:
            obj.write(str(doc))
        return path

    @staticmethod
    def _add_table(div, cols, data):
        table = div.table(klass="table table-condensed table-hover")
        tr = table.tr()
        for col in cols:
            tr.th(col)
        for datum in data:
            tr = table.tr()
            for thing in datum:
                tr.td(str(thing))

    @staticmethod
    def _add_figure(div, alttext, path, width=1000, **kwargs):
        div.img(src=os.path.basename(path), alt=alttext, width=str(width), **dict((key,str(val)) for key, val in kwargs.items()))

    def html_head_body(self):
        doc = HTML(newlines=True)
        doc.raw_text('<!DOCTYPE html>')
        htmldoc = doc.html(lang='en')
        head = htmldoc.head()

        ### set up header for bootstrap
        head.meta(charset='utf-8')
        head.meta(content='IE=edge')._attrs.update({'http-equiv':"X-UA-Compatible"}) ### forced into modifying _attr by "-" in attribute name
        head.meta(name="viewport", content="width=device-width, initial-scale=1")

        ### set up header for this specific template
        head.link(
            href="https://stackpath.bootstrapcdn.com/bootswatch/4.3.1/flatly/bootstrap.min.css",
            rel="stylesheet",
        )

        ### other header information
        head.meta(name="description", content=__process_name__+" report for %s (%s) within [%.3f, %.3f)"%(self.instrument, self.tag, self.start, self.end))
        head.meta(name="author", content=getpass.getuser()) ### whoever ran this is the author

        #----------------
        ### build body
        body = htmldoc.body()
        body = body.div(klass='container')

        return doc, head, body
